import glob
import numpy as np
import os
import shutil
import pandas as pd
import pyarrow as pa
import pathlib
from pyarrow.lib import Schema
from multiprocessing import Pool
from pyfortracc.utilities.utils import get_loading_bar, set_nworkers, check_operational_system
from pyfortracc.default_parameters import default_parameters



def concat(name_list, mode='track', clean=True, parallel=True):
    ''' 
    This function concatenates the results of the FortraCC algorithm
    into a single parquet file. The results are stored in the output_path/
    trackingtable if the processing is track or directory.
    The parquet file contains the following columns:

    Parameters
    -------
    name_list : dict
        Dictionary with the parameters to be used.
    mode : str
        Mode of the processing. Default is 'track'.
    clean : bool
        If True, the processing directory is removed after concatenation.
        Default is True.

    Returns
    -------
    None
    '''
    print('Concatenating:')
    proc_path = name_list['output_path'] + mode + '/processing/'
    # Set default parameters
    name_list = default_parameters(name_list)
    # Check operational system
    name_list, parallel = check_operational_system(name_list, parallel)
    # Set name of the output file
    if 'concat_path' not in name_list:
        name_list['concat_path'] = name_list['output_path']
    if mode == 'track':
        output_path = name_list['concat_path'] + 'track/trackingtable/'
    elif mode == 'forecast':
        output_path = name_list['concat_path'] + 'forecast/forecastable/'
    else:
        raise ValueError('Invalid mode')
    # Check if proc_path exists
    if not os.path.exists(proc_path) or os.path.isfile(output_path):
        print('No files to concatenate, the trackingtable already exists')
        print(output_path)
        return
    # Feature files
    fet_files = sorted(glob.glob(proc_path + 'features/*.parquet'))
    spt_files = sorted(glob.glob(proc_path + 'spatial/*.parquet'))
    lnk_files = sorted(glob.glob(proc_path + 'linked/*.parquet'))
    # Set default columns
    if name_list['default_columns']:
        default_cols = default_columns(name_list)
        if name_list['validation_scores']:
            last_spt = pd.read_parquet(spt_files[-1])
            concat_cols = last_spt.columns
            for col in concat_cols:
                if 'hit' in col or 'false-alarm' in col:
                    default_cols.append(col)
    # Create output directory
    if not os.path.exists(output_path):
        os.makedirs(output_path)
    # Get schema from the last file        
    schema = read_files((fet_files[-1], spt_files[-1], lnk_files[-1], 
                        default_cols, output_path, None, False))
    # Loading bar
    loading_bar = get_loading_bar(fet_files)
    if parallel:
        # Set number of workers
        n_workers = set_nworkers(name_list)
        with Pool(n_workers) as pool:
            for _ in pool.imap_unordered(read_files,
                                        [(fet_file, spt_file, lnk_file, 
                                            default_cols, output_path, schema, clean)
                                        for fet_file, spt_file, lnk_file in zip(fet_files, spt_files, lnk_files)]):
                loading_bar.update(1)
        pool.close()
    else:
        for fet_file, spt_file, lnk_file in zip(fet_files, spt_files, lnk_files):
            read_files((fet_file, spt_file, lnk_file, default_cols, 
                        output_path, schema, clean))
            loading_bar.update(1)
    loading_bar.close()           
    # Clean the processing directory
    if clean:
        shutil.rmtree(proc_path)
        
def read_files(args):
    ''' 
    This function reads multiple Parquet files, processes the data by dropping unnecessary columns, and concatenates them into a single DataFrame. 
    The resulting DataFrame is then filtered based on a set of default columns, indexed, and cleaned of duplicate columns.

    Parameters
    -------
    - `fet_file` : str
        Path to the feature file.
    - `spat_file` : str
        Path to the spatial file.
    - `link_file` : str
        Path to the link file.
    - `default_columns` : list
        List of columns to retain after concatenation.
    - `output_path` : str
        Directory path to save the processed Parquet file.
    - `schema` : Schema or None
        Schema for saving the output Parquet file. If None, the schema is inferred from the DataFrame.
    - `clean` : bool
        Flag to indicate whether to clean temporary files after saving.

    Returns
    -------
    None
    '''
    fet_file, spat_file, link_file, default_columns, output_path, schema, clean = args
    # Read feature files
    feat_df = pd.read_parquet(fet_file)
    spat_df = pd.read_parquet(spat_file)
    # Drop column threshold_level and trajectory from spat_df
    spat_df = spat_df.drop(columns=['threshold_level', 'trajectory'])
    link_df = pd.read_parquet(link_file)
    # Drop column threshold_level from link_df
    link_df = link_df.drop(columns=['threshold_level'])
    concat_df = pd.concat([feat_df, spat_df, link_df], axis=1)
    # Select columns
    concat_df = concat_df[default_columns]
    # Set index
    concat_df.set_index('cindex', inplace=True)
    # Drop duplicates columns
    concat_df = concat_df.loc[:,~concat_df.columns.duplicated()]
    # Check if schema is None
    if schema is None:
        return Schema.from_pandas(concat_df)
    # Get the name of file
    file_name = pathlib.Path(fet_file).name
    file_name = output_path + file_name
    # Check if file exist and remove it
    if os.path.exists(file_name):
        os.remove(file_name)
    # Save the parquet file
    save_parquet(concat_df, file_name, schema,
                clean=False, del_files=[fet_file, spat_file, link_file])
    

def save_parquet(concat_df, output_file, schema, clean=False, del_files=[]):
    ''' 
    This function saves a DataFrame to a Parquet file with specified schema and compression. 
    It also provides an option to delete input files after saving.

    Parameters
    -------
    - `concat_df` : pd.DataFrame
        The DataFrame to be saved as a Parquet file.
    - `output_file` : str
        Path where the output Parquet file will be saved.
    - `schema` : pyarrow.Schema
        Schema to enforce on the DataFrame before saving.
    - `clean` : bool, optional
        If True, deletes the original input files specified in `del_files` after saving. Default is False.
    - `del_files` : list, optional
        List of file paths to be deleted if `clean` is set to True. Default is an empty list.

    Returns
    -------
    None
    '''
    # table = pa.Table.from_pandas(concat_df, schema=schema)
    # df_converted = table.to_pandas()
    concat_df.to_parquet(output_file, engine='pyarrow', compression='gzip',
                            )
    if clean:
        for file in del_files:
            os.remove(file)

def default_columns(name_list=None):
    '''
    This function generates and returns a list of default columns based on the provided configuration in `name_list`.
    It dynamically adjusts the columns based on the presence of specific keys and their corresponding values in `name_list`.

    Parameters
    -------
    - `name_list` : dict 
        A dictionary containing configuration flags that determine which additional columns should be included. 
        The keys include 'thresholds', 'validation', 'validation_scores', 'spl_correction', 'mrg_correction', 
        and 'opt_correction', which control whether specific sets of columns should be added.

    Returns
    -------
    - `columns` : list
        A list of column names to be saved, which may include default columns and additional ones based on 
    '''
    columns = ['timestamp',
                'cindex',
                'uid',
                'threshold_level',
                'threshold',
                'status',
                'size',
                'lifetime',
                'expansion',
                'min',
                'mean',
                'max',
                'std',
                'u_',
                'v_',
                'inside_clusters',
                'board',
                'cluster_id',
                'file',
                'array_values',
                'array_y',
                'array_x',
                'vector_field',
                'trajectory',
                'geometry',
                'cluster_id',
                'past_idx',
                'merge_idx',
                'split_pr_idx']
    
    if len(name_list['thresholds']) > 1:
        # columns = columns + ['iuid']
        # Get position of uid column and add iuid after it
        uid_pos = columns.index('uid')
        columns = columns[:uid_pos + 1] + ['iuid'] + columns[uid_pos + 1:]
    if name_list['validation']:
        columns = columns + ['far', 'method']
    if name_list['validation_scores']:
        columns = columns + ['u_noc','v_noc',
                            'far_', 'hit_', 'false-alarm_']
    if name_list['spl_correction']:
        columns = columns + ['u_spl'] + ['v_spl']
        if name_list['validation_scores']:
            columns = columns + ['far_spl','hit_spl', 'false-alarm_spl']
    if name_list['mrg_correction']:
        columns = columns + ['u_mrg'] + ['v_mrg']
        if name_list['validation_scores']:
            columns = columns + ['far_mrg','hit_mrg', 'false-alarm_mrg']
    if name_list['validation_scores']:
        columns = columns + ['u_inc'] + ['v_inc']
        if name_list['validation_scores']:
            columns = columns + ['far_inc', 'hit_inc', 'false-alarm_inc']
    if name_list['opt_correction']:
        columns = columns + ['u_opt'] + ['v_opt']      
        if name_list['validation_scores']:
            columns = columns + ['far_opt', 'hit_opt', 'false-alarm_opt']
    if name_list['elp_correction']:
        columns = columns + ['u_elp'] + ['v_elp']
        if name_list['validation_scores']:
            columns = columns + ['far_elp', 'hit_elp', 'false-alarm_elp']
    if name_list['calc_dir']:
        columns = columns + ['dir']
    if name_list['calc_speed']:
        columns = columns + ['speed']
    if name_list['prv_uid']:
        columns = columns + ['prv_mrg_uids', 'prv_mrg_iuids', 
                             'prv_spl_uid', 'prv_spl_iuid']
    return columns



from shapely.geometry import Point
from shapely.affinity import scale
from pyfortracc.utilities.math_utils import uv_components

def ellipse_mtd(cur_df, prv_df):
    ''' 
    This method receive only events of Ellipse and create a new vector for ellipse fitting.
    The vector as created between the centroid of previous ellipse, and the centroid of current new ellipse.
    
    Parameters
    ----------
    cur_df : DataFrame
        current frame
    prev_df : DataFrame
        previous frame
        
    Returns
    ----------
    u_ : list 
        list of zonal (u) components
    v_ : list
        list of meridional (v) components
    '''  
    
    u_, v_ = [], []
    cur_df['elipse'] = cur_df.geometry.apply(envole_ellipse)
    prv_df['elipse'] = prv_df.geometry.apply(envole_ellipse)
    for _, row in cur_df.iterrows():
        cur_ellip_cent = row.elipse.centroid
        prv_ellip_cent = prv_df.loc[row['past_idx']].elipse.centroid
        # If prv_ellip_cent is more than one, get centroid between points
        if isinstance(prv_ellip_cent, Point) == False:
            prv_ellip_cent = prv_ellip_cent.unary_union.centroid
        uv_ = uv_components(prv_ellip_cent.coords[0], cur_ellip_cent.coords[0])
        u_.append(uv_[0])
        v_.append(uv_[1])
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # cur_df.geometry.boundary.plot(ax=ax, color='red')
    # prv_df.geometry.boundary.plot(ax=ax, color='blue')
    # cur_df.elipse.boundary.plot(ax=ax, color='green')
    # prv_df.elipse.boundary.plot(ax=ax, color='green')
    # plt.show()
    # exit()
    return u_, v_


def envole_ellipse(geometry, scale_factor=1):
    centroide = geometry.centroid
    minx, miny, maxx, maxy = geometry.bounds
    semi_eixo_x = (maxx - minx) / 2
    semi_eixo_y = (maxy - miny) / 2
    elipse = Point(centroide).buffer(1)
    elipse = scale(elipse, xfact=semi_eixo_x * scale_factor, yfact=semi_eixo_y * scale_factor)
    return elipse


'''
The choice of strategy for increasing the standard deviation of your matrix 
depends on the specific goals and constraints of your application. 
There isn't a one-size-fits-all "best" strategy because it depends on what
you're trying to achieve. Here are some common strategies, each with its
own advantages and use cases:

    Gaussian Noise Addition:
        Advantages: This approach is statistically rigorous and can precisely
        control the standard deviation of the resulting matrix.
        Use Case: If you need to simulate data with a specific level of noise 
        for testing or modeling purposes, adding Gaussian noise is a
        good choice.

    Smoothing or Blurring:
        Advantages: Applying a smoothing filter like Gaussian blur can 
        effectively increase the standard deviation of a matrix while preserving
        some spatial characteristics.
        Use Case: When working with images or spatial data and you want to 
        reduce high-frequency noise while increasing overall variability.

    Scaling or Amplification:
        Advantages: You can multiply the entire matrix by a scalar factor to
        increase its standard deviation. This is a straightforward method.
        Use Case: When you need to quickly adjust the standard deviation 
        without introducing randomness or altering the data distribution.

    Data Transformation:
        Advantages: For specific types of data, you may use mathematical 
        transformations to adjust the standard deviation. For example, 
        for normally distributed data, you can use the Z-score transformation.
        Use Case: When working with data that follows a known distribution and 
        you want to standardize it.

The "best" strategy depends on the nature of your data, your objectives, and 
the specific context of your application. Here are some considerations to help
you choose:

    Preservation of Data Characteristics: Consider whether you need to preserve 
    the original data characteristics, such as the data distribution or spatial
    features. Some methods (e.g., Gaussian noise addition) may introduce 
    randomness that alters the data distribution.

    Application-Specific Requirements: Think about the requirements of your 
    application. For example, if you're working on image processing, blurring 
    might be suitable for noise reduction and increasing variability.

    Control and Reproducibility: If you require precise control over the 
    standard deviation and need to reproduce the same results, methods like 
    Gaussian noise addition or scaling may be better suited.

    Data Understanding: Understand the underlying data and the effect of each 
    method on the data's interpretation and analysis. Consider how each approach
    may impact the quality of your results.

Ultimately, the choice of strategy should align with your specific objectives 
and the nature of your data. It may also involve experimentation to determine 
the best approach based on the results you achieve.
'''

import cv2
import numpy as np

# Function to add Gaussian noise to an image
def add_gaussian_noise(image, mean=0, stddev=25):
    """ 
    Add Gaussian noise to an image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to which Gaussian noise will be added. This should be a grayscale or color image in the form.
    mean : float, optional
        The mean of the Gaussian noise distribution. Default is 0.
    stddev : float, optional
        The standard deviation of the Gaussian noise distribution. Default is 25.

    Returns
    -------
    noisy_image : numpy.ndarray
        The image with added Gaussian noise. It has the same shape as the input image.

    Notes
    -----
    - Gaussian noise is generated using a normal distribution with the specified mean and standard deviation.
    - The noise is added to the image using 'cv2.add' to ensure proper handling of pixel values, preventing overflow or underflow.
    - The noise is cast to an unsigned 8-bit integer (np.uint8) before addition to match typical image data types.
    """
    # Generate Gaussian noise
    noise = np.random.normal(mean, stddev, image.shape).astype(np.uint8)
    # Add noise to the image
    noisy_image = cv2.add(image, noise)
    return noisy_image

# Function to apply Gaussian blur (smoothing) to an image
def apply_gaussian_blur(image, kernel_size=(5, 5), sigma=1.0):
    """ 
    Apply Gaussian blur (smoothing) to an image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to which Gaussian blur will be applied. This should be a grayscale or color image in the form.
    kernel_size : tuple of int, optional
        The size of the Gaussian kernel. It should be a tuple of two odd integers (width, height). Default is (5, 5).
    sigma : float, optional
        The standard deviation of the Gaussian kernel in both X and Y directions. Default is 1.0.

    Returns
    -------
    blurred_image : numpy.ndarray
        The image after applying Gaussian blur. It has the same shape as the input image.

    Notes
    -----
    - Gaussian blur is used to reduce image noise and detail by averaging pixel values with a Gaussian-weighted kernel.
    - The 'kernel_size' should be odd and positive; otherwise, the function will adjust the values automatically.
    - A larger 'sigma' value results in more significant blurring. If 'sigma' is set to 0, it is calculated based on 'kernel_size'.
    """
    # Apply Gaussian blur
    blurred_image = cv2.GaussianBlur(image, kernel_size, sigma)
    return blurred_image

# Function to scale the intensity of an image
def scale_image_intensity(image, scale_factor=2.0):
    """ 
    Scale the intensity of an image by a specified factor.

    Parameters
    ----------
    image : numpy.ndarray
        The input image whose intensity will be scaled. This should be a grayscale or color image.
    scale_factor : float, optional
        The factor by which to scale the image's intensity. Default is 2.0.

    Returns
    -------
    scaled_image : numpy.ndarray
        The image with scaled intensity. It has the same shape as the input image.

    Notes
    -----
    - Intensity scaling multiplies each pixel value by the 'scale_factor'. This can brighten or darken the image.
    - The function uses 'cv2.multiply' to handle the scaling, which ensures that pixel values stay within the valid range (0-255) for 8-bit images.
    - If 'scale_factor' is greater than 1.0, the image will become brighter. If it's between 0 and 1.0, the image will become darker.
    - Scaling may cause pixel values to exceed the maximum value of 255, leading to clipping.
    """
    # Scale the image intensity
    scaled_image = cv2.multiply(image, np.array([scale_factor]))
    return scaled_image

# Function to apply Z-score transformation to an image
def zscore_transform(image):
    """ 
    Apply Z-score transformation to an image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to be transformed, typically in the form of a NumPy array.

    Returns
    -------
    zscore_image : numpy.ndarray
        The image after Z-score transformation, with the same shape as the input image. The output values will be standardized, having a mean of 0 and a standard deviation of 1.

    Notes
    -----
    - The Z-score transformation standardizes the image by subtracting the mean and dividing by the standard deviation of the pixel values.
    - This transformation is useful for normalizing the image, especially in preprocessing steps for machine learning models.
    - The image is converted to `float32` before performing the transformation to ensure precision and to avoid integer overflow or underflow.
    - The resulting values may not be in the typical 0-255 range, as they represent standard deviations from the mean.
    """
    # Convert image to float32 for Z-score calculation
    image_float = image.astype(np.float32)
    # Calculate mean and standard deviation of the image
    mean = np.mean(image_float)
    stddev = np.std(image_float)
    # Apply Z-score transformation
    zscore_image = (image_float - mean) / stddev
    return zscore_image

# Function to perform histogram equalization
def histogram_equalization(image):
    """ 
    Perform histogram equalization on an image to enhance contrast.

    Parameters
    ----------
    image : numpy.ndarray
        The input image for which histogram equalization will be applied.

    Returns
    -------
    equalized_image : numpy.ndarray
        The image after applying histogram equalization. If the input was a color image, the output will be a grayscale image.

    Notes
    -----
    - Histogram equalization is a technique for improving the contrast in images by spreading out the most frequent intensity values.
    - If the input image is a color image, it will be converted to grayscale before applying histogram equalization.
    - The function uses 'cv2.equalizeHist', which applies the equalization on a per-channel basis for grayscale images.
    """
    # Convert image to grayscale if it's in color
    if len(image.shape) == 3:
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        gray_image = image
    # Apply histogram equalization
    equalized_image = cv2.equalizeHist(gray_image)
    return equalized_image

# Function to perform image thresholding
def threshold_image(image, threshold_value=128, max_value=255,
                    threshold_type=cv2.THRESH_BINARY):
    """ 
    Perform image thresholding to create a binary or segmented image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to which thresholding will be applied.
    threshold_value : int, optional
        The threshold value used to classify the pixel values. Default is 128.
    max_value : int, optional
        The maximum value assigned to pixels that meet the thresholding condition. Default is 255.
    threshold_type : int, optional
        The type of thresholding to apply. Default is 'cv2.THRESH_BINARY'.

    Returns
    -------
    thresholded_image : numpy.ndarray
        The image after applying the specified thresholding. If the input was a color image, the output will be a grayscale thresholded image.

    Notes
    -----
    - Thresholding is a technique used to create binary images by setting pixel values to a max value if they exceed a certain threshold, or to zero otherwise.
    - The function first converts color images to grayscale before applying thresholding, as thresholding is typically applied to single-channel images.
    - The 'threshold_type' parameter allows for different types of thresholding behaviors, providing flexibility in image segmentation.
    """
    # Convert image to grayscale if it's in color
    if len(image.shape) == 3:
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        gray_image = image
    # Apply thresholding
    _, thresholded_image = cv2.threshold(gray_image, threshold_value, max_value,
                                        threshold_type)
    return thresholded_image

# Function to sharpen an image
def sharpen_image(image):
    """ 
    Sharpen an image using a Laplacian kernel.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to be sharpened.

    Returns
    -------
    sharpened_image : numpy.ndarray
        The image after applying the sharpening filter, with enhanced edges and details.

    Notes
    -----
    - Sharpening an image enhances its edges and details, making the features more pronounced.
    - The sharpening kernel used here is a Laplacian kernel, which emphasizes the center pixel and subtracts the neighboring pixel values.
    - The function uses 'cv2.filter2D' to apply the convolution of the image with the sharpening kernel. This method is effective for both grayscale and color images.
    - The sharpening process may introduce noise or artifacts, particularly in images with already high contrast.
    """
    # Define a sharpening kernel (Laplacian)
    kernel = np.array([[0, -1, 0],
                    [-1, 5,-1],
                    [0, -1, 0]], dtype=np.float32)
    
    # Apply convolution with the sharpening kernel
    sharpened_image = cv2.filter2D(image, -1, kernel)
    return sharpened_image

# Function to perform median filtering for denoising
def denoise_median(image, kernel_size=5):
    """ 
    Perform median filtering to denoise an image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to be denoised.
    kernel_size : int, optional
        The size of the kernel to be used for median filtering. It must be a positive odd integer (e.g., 3, 5, 7). The default value is 5.

    Returns
    -------
    denoised_image : numpy.ndarray
        The image after applying median filtering, with reduced noise and preserved edges.

    Notes
    -----
    - Median filtering is a non-linear filtering technique commonly used to remove salt-and-pepper noise while preserving edges.
    - The `kernel_size` parameter determines the size of the neighborhood considered around each pixel when applying the median filter. A larger kernel size will result in stronger noise reduction but may also cause more blurring.
    - The function uses 'cv2.medianBlur' to perform the median filtering, which is effective for both grayscale and color images.
    - The kernel size must be an odd number to ensure that the median can be correctly computed from the pixel neighborhood.
    """
    # Apply median filtering
    denoised_image = cv2.medianBlur(image, kernel_size)
    return denoised_image

# Function to perform Canny edge detection
def detect_edges_canny(image, low_threshold=50, high_threshold=150):
    """ 
    Perform Canny edge detection on an image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image on which edge detection will be performed.
    low_threshold : int, optional
        The lower bound for the hysteresis thresholding. Default is 50.
    high_threshold : int, optional
        The upper bound for the hysteresis thresholding. Default is 150.

    Returns
    -------
    edges : numpy.ndarray
        The binary image showing the detected edges.

    Notes
    -----
    - Canny edge detection is a multi-stage algorithm that identifies the edges in an image by looking for areas of rapid intensity change.
    - The 'low_threshold' and 'high_threshold' parameters control the sensitivity of the edge detection. Pixels with gradient values below the 'low_threshold' are discarded, while those above the 'high_threshold' are considered strong edges. 
    - The input image is automatically converted to grayscale if it is in color, as the Canny algorithm works on single-channel images.
    """
    # Convert image to grayscale if it's in color
    if len(image.shape) == 3:
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    else:
        gray_image = image
    # Apply Canny edge detection
    edges = cv2.Canny(gray_image, low_threshold, high_threshold)
    return edges

# Function to perform dilation operation
def dilate_image(image, kernel_size=3):
    """ 
    Perform dilation operation on an image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to be dilated.
    kernel_size : int, optional
        The size of the square structuring element (kernel) used for dilation. Default is 3.

    Returns
    -------
    dilated_image : numpy.ndarray
        The image after applying the dilation operation.

    Notes
    -----
    - Dilation is a morphological operation that enlarges the boundaries of objects in an image. It is commonly used to increase the size of foreground objects, fill in small holes, or connect disjointed elements.
    - The 'kernel_size' parameter defines the size of the structuring element used for dilation. A larger kernel size results in more significant dilation.
    - The structuring element (kernel) is generated as a square matrix of ones using 'np.ones' with the specified 'kernel_size'.
    - The function applies the dilation operation using 'cv2.dilate', which shifts the kernel over the image and replaces the center pixel with the maximum pixel value covered by the kernel.
    """
    kernel = np.ones((kernel_size, kernel_size), np.uint8)
    dilated_image = cv2.dilate(image, kernel, iterations=1)
    return dilated_image

# Function to perform erosion operation
def erode_image(image, kernel_size=3):
    """
    Perform erosion operation on an image.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to be eroded.
    kernel_size : int, optional
        The size of the square structuring element (kernel) used for erosion. Default is 3.

    Returns
    -------
    eroded_image : numpy.ndarray
        The image after applying the erosion operation.

    Notes
    -----
    - Erosion is a morphological operation that reduces the boundaries of objects in an image. It is often used to remove small-scale noise, separate connected components, or shrink foreground objects.
    - The 'kernel_size' parameter defines the size of the structuring element used for erosion. A larger kernel size will result in more significant erosion.
    - The structuring element (kernel) is created as a square matrix of ones using 'np.ones' with the specified 'kernel_size'.
    - The function performs erosion using 'cv2.erode', which shifts the kernel over the image and replaces the center pixel with the minimum pixel value covered by the kernel.
    """
    kernel = np.ones((kernel_size, kernel_size), np.uint8)
    eroded_image = cv2.erode(image, kernel, iterations=1)
    return eroded_image

# Function to perform texture analysis (example: Gabor filter)
def texture_analysis(image, kernel_size=7, theta=0, sigma=2.0, frequency=0.2):
    """ 
    Perform texture analysis on an image using a Gabor filter.

    Parameters
    ----------
    image : numpy.ndarray
        The input image to which the Gabor filter will be applied.
    kernel_size : int, optional
        The size of the Gabor kernel. The kernel will be a square matrix of size (kernel_size, kernel_size). Default is 7.
    theta : float, optional
        The orientation of the Gabor filter in radians. The angle is measured from the x-axis. Default is 0 (horizontal orientation).
    sigma : float, optional
        The standard deviation of the Gaussian envelope of the Gabor filter. Default is 2.0.
    frequency : float, optional
        The spatial frequency of the sinusoidal component of the Gabor filter. Default is 0.2.

    Returns
    -------
    filtered_image : numpy.ndarray
        The image after applying the Gabor filter, highlighting texture features.

    Notes
    -----
    - A Gabor filter is used for texture analysis and edge detection. It is especially effective at capturing texture patterns and frequencies.
    - The 'kernel_size' determines the dimensions of the Gabor kernel, which will be a square matrix with size (kernel_size, kernel_size).
    - The 'theta' parameter controls the orientation of the filter. A value of 0 indicates horizontal orientation, while other values rotate the filter accordingly.
    - The 'sigma' parameter controls the width of the Gaussian envelope, affecting the filter's scale and the extent of smoothing.
    - The 'frequency' parameter defines the spatial frequency of the sinusoidal component, which influences the filter's sensitivity to different texture patterns.
    - The function uses OpenCV's 'cv2.getGaborKernel' to create the Gabor kernel and 'cv2.filter2D' to apply the filter to the image.
    """
    # Create a Gabor kernel
    kernel = cv2.getGaborKernel((kernel_size, kernel_size), sigma, theta, 1.0, 
                                frequency, 0, ktype=cv2.CV_32F)
    # Apply Gabor filter to the image
    filtered_image = cv2.filter2D(image, cv2.CV_8UC3, kernel)
    return filtered_image



import numpy as np

def dbz2mm6m3(dBZ_array):
    """ 
    Convert radar reflectivity (dBZ) to the unit of mm^6/m^3.

    Parameters
    ----------
    dBZ_array : array-like
        An array of reflectivity values in dBZ (decibels of Z), which is a logarithmic measure of radar reflectivity.

    Returns
    -------
    mm6m3_values : array-like
        The converted reflectivity values in mm^6/m^3, representing the volume concentration of hydrometeors.

    Notes
    -----
    - The conversion formula used is: mm^6/m^3 = 10^(dBZ/10)
    - This conversion assumes that dBZ is a logarithmic measurement of the reflectivity factor (Z) in mm^6/m^3.
    """
    mm6m3_values = 10.0 ** (dBZ_array / 10.0) # Convert reflectivity to mm^6/m^3
    return mm6m3_values


def mm6m32dbz(mm6m3_array):
    """ 
    Convert volume concentration of hydrometeors from mm^6/m^3 to radar reflectivity in dBZ.

    Parameters
    ----------
    mm6m3_array : array-like
        An array of reflectivity values in units of mm^6/m^3, representing the volume concentration of hydrometeors.

    Returns
    -------
    dBZ_array : array-like
        The converted reflectivity values in decibels of Z (dBZ), which is a logarithmic measure of radar reflectivity.

    Notes
    -----
    - The conversion formula used is: dBZ = 10 * log10(mm^6/m^3)
    - This conversion assumes that the input values are in mm^6/m^3, and converts them to the logarithmic dBZ scale used in radar meteorology.
    """
    dBZ_array = 10.0 * np.log10(mm6m3_array) # Convert mm^6/m^3 to dBZ
    return dBZ_array


def dbz2mmh(dBZ_array, a=200.0, b=1.6):
    """ 
    Convert radar reflectivity in dBZ to rainfall rate in mm/h using the Marshall-Palmer formula.

    Parameters
    ----------
    dBZ_array : array-like
        An array of reflectivity values in decibels (dBZ). This represents radar reflectivity on a logarithmic scale.
    a : float, optional
        The scaling factor in the Marshall-Palmer formula. Default is 200.
    b : float, optional
        The exponent in the Marshall-Palmer formula. Default is 1.6.

    Returns
    -------
    R : array-like
        The calculated rainfall rate in millimeters per hour (mm/h).

    Notes
    -----
    - The Marshall-Palmer formula used is: R = (Z / a)^(1/b), where Z is the reflectivity factor in mm^6/m^3.
    - The conversion assumes that the dBZ values are converted to linear reflectivity first, and then the Marshall-Palmer relationship is applied to estimate the rainfall rate.
    """
    Z = 10.0 ** (dBZ_array / 10.0) # Convert reflectivity to mm^6/m^3
    R = (Z / a) ** (1.0 / b) # Apply Marshall-Palmer formula
    return R


def mmh2dbz(mmh_array, a=200.0, b=1.6):
    """ 
    Convert rainfall rate in mm/h to radar reflectivity in dBZ using the Marshall-Palmer formula.

    Parameters
    ----------
    mmh_array : array-like
        An array of rainfall rates in millimeters per hour (mm/h).
    a : float, optional
        The scaling factor used in the Marshall-Palmer formula. Default is 200.
    b : float, optional
        The exponent used in the Marshall-Palmer formula. Default is 1.6.

    Returns
    -------
    dBZ_array : array-like
        The converted reflectivity values in decibels (dBZ), which is a logarithmic measure of radar reflectivity.

    Notes
    -----
    - The Marshall-Palmer formula used is: Z = a * R^b, where Z is the reflectivity factor in mm^6/m^3 and R is the rainfall rate in mm/h.
    - After computing the reflectivity factor Z, the formula converts it to dBZ using the equation: dBZ = 10 * log10(Z).
    - The conversion assumes that the input values are in mm/h and converts them to the logarithmic dBZ scale used in radar meteorology.
    """
    Z = a * mmh_array**b # Convert mm/h to mm^6/m^3
    dBZ_array = 10.0 * np.log10(Z) # Convert mm^6/m^3 to dBZ
    return dBZ_array